function [s, max_res] = DS_like_shift_search(V, S, A, k, smin, smax, complex, szB, mode )
% DS_LIKE_SHIFT_SEARCH  -  Does a shift search similar to what is presented
% in V. Druskin and V. Simoncini, "Adaptive rational Krylov subspace for
% large-scale dynamical systems", Systems Control Lett., 2011.
% Implements the ideas of equations (2.2) and (2.4).
%
%  Targets generalized Lyapunov equations: AX + XA^T + NXN^T + BB^T = 0,
%  but not restricted to that. Finds shifts for spaces of the type
%  {B, (A-S(1))^{-1}w_1, (A-S(2))^{-1}w_2, ... , (A-S(k))^{-1}w_k}
%
%  [s, max_res] = DS_like_shift_search(V, S, A, k, smin, smax, complex, szB, mode)
%
%  Input:
%  V - The current basis
%  S - The previously used shifts to generate V
%  A - The matrix used in rational Krylov (A-sI)^{-1}
%  k - The number of outer iterations of the space already computed
%  smin - Estimate of the smallest eigenvalue of A
%  smax - Estimate of the largest eigenvalue of A
%  complex - Set to 1 if complex shifts are desired, defult is 0
%  szB - The rank of the right-hand-side B.
%  mode - Descides what mode to run in. Available options are:
%    * 'classic' - Described in the article above. Does not take the szB
%                  first columns of V into account since they are assumed
%                  to be an orthogonalized version of B. Assumes space V is
%                  expanded with szB columns in each iteration, typicall
%                  then w_i = B.
%    * 'useB' - Uses all of V when computing the Ritz values. Moreover
%               assumes that space V is only expanded with one column in 
%               each iteration, i.e. w_i is a vector.
%               
%    * 'notB' - Hybrid between the two above. Does not take the szB first
%               columns of V into account since they are assumed to be an
%               orthogonalized version of B. Assumes space is expanded with
%               one column in each iteration, i.e., w_i is a vector.
%
%  Output:
%  s - The found shift
%  max_res - The maximum residual at where the shift was found
% 

    s = NaN;
    max_res = NaN;
    
    if ~exist('complex', 'var') % If not specified search for real shifts
      complex = 0;
    end
    
    if k == 1
      s = smax;
    elseif (k == 2) || (k==3 && imag(smax)~=0)
      s = smin;
    else
      if ~exist('szB', 'var') % If rank of B is not given then do classical DS
        mode = 'useB';
      end
      if strcmp(mode,'classic')
        uno = ones(szB,1);
        SS = kron(S(1:(k-1)),uno);
        VV = V(:,szB+1:end);
      elseif strcmp(mode,'useB') % Only 1 shift used and use all of the space
        SS = S(1:(k-1));
        VV = V;
      elseif strcmp(mode,'notB') % Only 1 shift used and ignore B-part of space
        SS = S(1:(k-1));
        VV = V(:,szB+1:end);
      else
        error('Invalid mode given')
      end
      Ap = VV'*A*VV;
      Ritz_vals = sort(eig(full(Ap)));
      if (strcmp(mode,'classic') || strcmp(mode,'notB')) && (length(Ritz_vals) ~= length(SS))
        warning('Space does not grow. Returning early! k = %i', k)
        return
      end
      
      intervals_to_search = sort([-Ritz_vals.', smin, smax]);
      intervals_to_search = intervals_to_search(intervals_to_search >= eps*10);
      if (complex == 1) && any(imag(Ritz_vals) & real(Ritz_vals)<-eps*10) && length(Ritz_vals)>2
        % Accept complex poles, have complex Ritz values, have more than two points
        conv_hull_idx = convhull(real(intervals_to_search), imag(intervals_to_search));
        intervals_to_search = intervals_to_search(conv_hull_idx);
      else
        intervals_to_search = real(intervals_to_search);
      end
      max_res = -inf;

      for i = 1:length(intervals_to_search)-1
        for ss = linspace(intervals_to_search(i), intervals_to_search(i+1),200)
          if (strcmp(mode,'classic') || strcmp(mode,'notB'))
            abs_r_inv = abs(prod((ss-SS)./(ss-Ritz_vals)));
          else
            abs_r_inv = abs(prod((ss-SS))/prod((ss-Ritz_vals)));
          end
          if(abs_r_inv > max_res)
            max_res = abs_r_inv;
            s = ss;
          end
        end
      end
      
    end
    
    if real(s)<0, s=-real(s)+1i*imag(s), end  %safeguard strategy

end
