function [ Y ] = inner_solver_GenLyap( A, N, B, V )
%INNER_SOLVER_GENLYAP  -  Module for solving the projected problem.
%  Aims at computing the Galerkin projection of the generalized Lyapunov
%  equation,
%  AX + XA^T + \sum_{i=1}^m N{i}XN{i}^T + BB^T = 0,
%  projected on the subspace spanned by V (assumed to be Orthogonal).
%
%  OBS: Allowed to be optimized depending on problem structure!
%
%  [ Y ] = inner_solver_GenLyap( A, N, B, V)
%
%  Input:
%  A - Matrix defining the Lyapunov operator
%  N - Cell array containing the "generalized" terms of the equation
%  B - Low-rank factor of the "right-hand-side"
%  V - Orthogonal matrix containing a basis of the subpsace projected on
%
%  Output:
%  Y - The soltion to the projected problem (OBS: the small-scale problem!)
%

  if ~iscell(N); error('N needs to be a cell-array'); end;
  
  Ap = V'*A*V;
  Np = cell(length(N),1);
  for i = 1:length(N)
    Np{i} = V'*N{i}*V;
  end
  Bp = V'*B; % TODO: Can be optimized since orth(B) is the starting block
  
  % Obs: Solve with naive method to allow for most general application
  n = size(V,2);
  I = speye(n,n);
  L = kron(I,Ap) + kron(Ap,I);
  for i = 1:length(N)
    L = L + kron(Np{i},Np{i});
  end 
  rhs = (-Bp)*Bp';
  Y = reshape(L\rhs(:), n, n);

end

