function [V, S, max_res] = rat_Krylov_type_GenLyap_DS( A, N, B, m, smin, smax, update_n_step, complex_shifts )
%RAT_KRYLOV_TYPE_GENLYAP_DS
%  Using a naive block-Arnoldi to compute a rational Krylov type subspace
%  aimed at approximating the solution to the generalized Lyapunov equation
%             AX + XA^T + \sum_{i=1}^m N{i}XN{i}^T + BB^T = 0.
%  
%  The space created is
%      V = {B, (A-S(1))^{-1}r_1, (A-S(2))^{-1}r_2, ... , (A-S(k))^{-1}r_k},
%  where r_i is the dominating left signular vector of the residual in the
%  previous space. Shifts S are computed in accordance with
%  V. Druskin and V. Simoncini, "Adaptive rational Krylov subspace for
%  large-scale dynamical systems", Systems Control Lett., 2011.
%  Using a call to DS_LIKE_SHIFT_SEARCH.
%  Returns a space of dimension size(B,2)+(m-1), that is, it does m-1 steps
%
%  [V, S, max_res] = rat_Krylov_type_GenLyap_DS( A, N, B, m, smin, smax, update_n_step, complex_shifts )
%
%  Input:
%  A - Matrix defining the Lyapunov operator
%  N - Cell array containing the "generalized" terms of the equation
%  B - Low-rank factor of the "right-hand-side"
%  m - The method does m-1 steps
%  smin - Estimate of the smallest eigenvalue of A
%  smax - Estimate of the largest eigenvalue of A
%  update_n_step - Allows to update the residual direction r only every
%    update_n_step of the outer iteration. Can be set to 1 if it ise
%    desired to update the residual direction at every iteration
%  complex_shifts - Determines if shifts are complex (1) or real (0). In
%    case of complex shifts, real arithmetic is still enforced in the basis
%
%  Output:
%  V - A matrix containing an orthogonal basis of the suggested search
%    space for approximating the generalized Lyapunov equation.
%  s - The used shifts
%  max_res - The maximum residuals at where the corresponding shifts were found
%

if ~iscell(N); error('N needs to be a cell-array'); end;

if smax < smin
  error('Input error: smax < smin')
end

n = size(A,1);
mm = m-1;
r = [];
R = [];

% V = zeros(n, r*m);
S = NaN((mm), 1);
max_res = -inf*ones((mm), 1);

if issparse(A)
  I = speye(n,n);
else
  I = eye(n,n);
end



V = qr_orth(B); %Orthogonalize and normalize within starting block

k = 0;
while k < (mm)
  k = k + 1;
  fprintf('Iteration: %2.i of %i\n',k,(mm))
  
  [r, R] = update_residual_direction_GenLyap(A, N, B, V, k, update_n_step, r, R);
  
  % Find a new shift
  [s, mrs] = DS_like_shift_search(V, S, A, k, smin, smax, complex_shifts, [], 'useB');
  if(isnan(s) && isnan(mrs)); return; end %Abort if space is not growing
  
  % Update the basis
  if imag(s) == 0 % One real shift
    S(k) = s;
    max_res(k) = mrs;
    % Expand the basis
    Vi = (A - s*I)\r;
    V = qr_orth([V,Vi]);
  elseif imag(s) ~= 0 % One complex conjugate shift
    S(k) = s;
    S(k+1) = conj(s);
    max_res(k) = mrs;
    max_res(k+1) = mrs;
    % Expand the basis
    Vi = (A - s*I)\r;
    Vi = [real(Vi), imag(Vi)];
    V = qr_orth([V,Vi]);
    k = k+1;
  else
    error('This should not be possible.')
  end
end

end
